#include<stdint.h>
#include<stddef.h>
#include"../sys.h"
#include"dio.h"

#define IO_BASE 0x1F0
#define REG_DATA 0
#define REG_RERROR_WFEATURES 1
#define REG_SECTOR_COUNT 2
#define REG_LBA_LOW 3
#define REG_LBA_MIDDLE 4
#define REG_LBA_HIGH 5
#define REG_DRIVE 6
#define REG_RSTATUS_WCOMMAND 7

char stack[768];
uint16_t data[256];

static void waitUntilReady() {
	while((inb(IO_BASE + REG_RSTATUS_WCOMMAND) & 128) == 128) {
	}
	while((inb(IO_BASE + REG_RSTATUS_WCOMMAND) & 8) == 0) {
	}
}

static void wait400() {
	inb(0x3F6);
	inb(0x3F6);
	inb(0x3F6);
	inb(0x3F6);
}

void startc() {
	outb(IO_BASE + REG_DRIVE, 0xA0);
	outb(IO_BASE + REG_SECTOR_COUNT, 0);
	outb(IO_BASE + REG_LBA_LOW, 0);
	outb(IO_BASE + REG_LBA_MIDDLE, 0);
	outb(IO_BASE + REG_LBA_HIGH, 0);
	outb(IO_BASE + REG_RSTATUS_WCOMMAND, 0xEC);
	
	waitUntilReady();
	
	/* IDENTIFY data, which we don't need (yet). */
	for(int i = 0; i < 256; i++) {
		inw(IO_BASE + REG_DATA);
	}
	
	uint32_t canalID = sys_canal_create(0x42444844, 0x00000000);
	
	uint16_t *buf = (void*) -1;
	do {
		buf = sys_canal_accept(canalID);
	} while(buf == (void*) -1);
	
	while(1) {
		memory_barrier();
		sys_signal_wait(SYS_SIGNAL_WAIT_ANY);
		
		LumaDIOCommand *cmd = (void*) buf;
		if(cmd->op == LUMA_DIO_COMMAND_READ) {
			uint8_t count = cmd->read.count;
			
			outb(IO_BASE + REG_DRIVE, 0xE0);
			outb(IO_BASE + REG_RERROR_WFEATURES, 0x00);
			outb(IO_BASE + REG_SECTOR_COUNT, count);
			outb(IO_BASE + REG_LBA_LOW, (cmd->read.lba >> 0) & 0xFF);
			outb(IO_BASE + REG_LBA_MIDDLE, (cmd->read.lba >> 8) & 0xFF);
			outb(IO_BASE + REG_LBA_HIGH, (cmd->read.lba >> 16) & 0xFF);
			outb(IO_BASE + REG_RSTATUS_WCOMMAND, 0x20);
			
			int w = 0;
			while(count--) {
				waitUntilReady();
				for(int i = 0; i < 256; i++) {
					buf[w++] = inw(IO_BASE + REG_DATA);
				}
			}
			
			wait400();
		}
		
		memory_barrier();
		sys_signal_send(buf);
	}
}