#include"../sys.h"

asm("modentry:\n"
"mov $stack + 1536, %esp\n"
"call startc");

#include"../std.h"

char stack[2048];

#define FORMAT_CHANNELS(n) ((n) - 1)
#define FORMAT_16BPS (1 << 4)
#define FORMAT_RATE_MUL_1 0
#define FORMAT_RATE_DIV_1 0
#define FORMAT_PCM (0 << 15)
#define FORMAT_48KHZ (0 << 14)

#define STANDARD_48KHZ_16BPS (FORMAT_CHANNELS(1) | FORMAT_PCM | FORMAT_48KHZ | FORMAT_RATE_MUL_1 | FORMAT_RATE_DIV_1 | FORMAT_16BPS)

#define STREAM_RUN_BIT 2
#define STREAM_NO(n) ((n) << 20)

struct BufDescEntry {
	uint64_t addr;
	uint32_t length;
	uint32_t flags;
};
static struct BufDescEntry *bufDesc;

volatile uint32_t *HDA;

#define WIDGET_TYPE(w) (((w)->cap >> 20) & 15)
#define WIDGET_AMP_OVERRIDE(w) (((w)->cap >> 3) & 1)
struct Widget {
	uint32_t cap;
	uint8_t connectionCount;
	uint8_t queued; //Used for BFS
	uint8_t parent;
};
static uint8_t WidgetCount, WidgetsStart;

static uint8_t TheCodec, ThePin;
static uint32_t DefaultAFGInAmpCaps;
static uint32_t DefaultAFGOutAmpCaps;

static struct Widget widgets[64];

static uint32_t cmd(uint8_t codec, uint8_t node, uint32_t c) {
	HDA[26] = 2;
	HDA[24] = (codec << 28) | (node << 20) | c;
	HDA[26] = 1;
	while((HDA[26] & 3) != 2);
	return HDA[25];
}
static uint32_t verb(uint8_t codec, uint8_t node, uint16_t v, uint8_t payload) {
	return cmd(codec, node, (v << 8) | payload);
}

static uint32_t verb16(uint8_t codec, uint8_t node, uint16_t v, uint16_t payload) {
	return cmd(codec, node, (v << 16) | payload);
}

static void unmute(uint8_t widget, int output, int from, int to) {
	uint32_t ampcap;
	if(WIDGET_AMP_OVERRIDE(widgets + widget - WidgetsStart)) {
		ampcap = verb(TheCodec, widget, 0xF00, 0xD + 5 * output);
	} else {
		ampcap = output ? DefaultAFGOutAmpCaps : DefaultAFGInAmpCaps;
	}
	for(int i = from; i <= to; i++) {
		verb16(TheCodec, widget, 0x3, (ampcap & 127) | (i << 8) | (3 << 12) | ((1 + output) << 14));
	}
}

static void klogc(int c) {
	asm volatile("outb %%al, $0xE9"::"a"(c):);
}
static void klogp(void *p) {
	static char hex[16]="0123456789ABCDEF";
	
	uintptr_t i = (uintptr_t) p;
	klogc(hex[(i >> 28) & 0xF]);
	klogc(hex[(i >> 24) & 0xF]);
	klogc(hex[(i >> 20) & 0xF]);
	klogc(hex[(i >> 16) & 0xF]);
	klogc(hex[(i >> 12) & 0xF]);
	klogc(hex[(i >> 8) & 0xF]);
	klogc(hex[(i >> 4) & 0xF]);
	klogc(hex[(i >> 0) & 0xF]);
}

void startc() {
	uint32_t hcId = sys_pci_claim(0x04030000);
	if(hcId == (uint32_t) -1) {
		while(1) sys_signal_wait(SYS_SIGNAL_WAIT_ANY, NULL);
	}
	
	// Clear TCSEL
	sys_pci_write_1(hcId, 0x44, sys_pci_read(hcId, 0x44) & ~7);
	
	// Enable busmastering
	sys_pci_write_2(hcId, 0x04, sys_pci_read(hcId, 0x04) | 7);
	
	HDA = sys_vpm_map(SYS_MAP_VIRT_ANY, (void*) (sys_pci_read(hcId, 0x10) & 0xFFFFFFF0), 4096);
	
	// Start reset.
	HDA[2] = 0;
	while((HDA[2] & 1) != 0);
	sys_sleep(50);
	
	// Finish reset.
	HDA[2] = 1;
	while((HDA[2] & 1) == 0);
	sys_sleep(10);
	
	int16_t *audbuf = sys_vpm_map(SYS_MAP_VIRT_ANY, SYS_MAP_PHYS_ANY, 8192 * sizeof(*audbuf));
	for(int i = 0; i < 8192; i++) {
		//// 2 buffers of square wave
		audbuf[i] = i % 128 < 64 ? 10000 : -10000;
		//audbuf[i] = 0;
	}
	
	bufDesc = sys_vpm_map(SYS_MAP_VIRT_ANY, SYS_MAP_PHYS_ANY, 4096);
	
	bufDesc[0].addr = (uintptr_t) sys_vpm_get_phys(&audbuf[0]);
	bufDesc[0].length = 8192;
	bufDesc[0].flags = 1;
	
	bufDesc[1].addr = (uintptr_t) sys_vpm_get_phys(&audbuf[4096]);
	bufDesc[1].length = 8192;
	bufDesc[1].flags = 1;
	
	uint32_t *dmapos = (uint32_t*) &bufDesc[8];
	HDA[28] = 1 | (uintptr_t) sys_vpm_get_phys(dmapos);
	HDA[29] = 0;
	
	uint16_t gcap = *(volatile uint16_t*) HDA;
	uint8_t inputStreamCount = (gcap >> 8) & 15;
	
	volatile uint32_t *outStream = HDA + 0x20 + 8 * inputStreamCount;
	outStream[0] &= ~2; // Disable run
	
	// Reset.
	outStream[0] |= 1;
	int timeout = 500;
	while((outStream[0] & 1) == 0 && timeout--);
	
	// Unreset.
	outStream[0] &= ~1;
	timeout = 500;
	while((outStream[0] & 1) != 0 && timeout--);
	
	outStream[0] = STREAM_NO(1);
	outStream[2] = 16384;
	((volatile uint16_t*) outStream)[9] = STANDARD_48KHZ_16BPS;
	((volatile uint16_t*) outStream)[6] = 1;
	outStream[6] = (uintptr_t) sys_vpm_get_phys(bufDesc);
	outStream[7] = 0;
	
	sys_sleep(10);
	
	uint16_t statests = ((volatile uint16_t*) HDA)[7];
	for(int codec = 0; codec < 15; codec++) {
		if(statests & (1 << codec)) {
			TheCodec = codec;
			break;
		}
	}
	
	// Get subordinate nodes
	uint32_t boob = verb(TheCodec, 0, 0xF00, 4);
	
	uint8_t startIdx = (boob >> 16) & 0xFF;
	uint8_t endIdx = startIdx + (boob & 0xFF);
	for(int idx = startIdx; idx < endIdx; idx++) {
		// Get function group type
		uint32_t type = verb(TheCodec, idx, 0xF00, 5);
		
		if((type & 127) == 1) {
			// Is audio function group
			
			// Power
			verb(TheCodec, idx, 0x705, 0);
			
			sys_sleep(200);
			
			DefaultAFGInAmpCaps = verb(TheCodec, idx, 0xF00, 0xD);
			DefaultAFGOutAmpCaps = verb(TheCodec, idx, 0xF00, 0x12);
			
			// Get subordinate nodes
			uint32_t tit = verb(TheCodec, idx, 0xF00, 4);
			
			// Iterate over widgets
			uint8_t wIdxStart = (tit >> 16) & 0xFF;
			uint8_t wIdxEnd = wIdxStart + (tit & 0xFF);
			
			int outPinWPriority = 0;
			int outPinWIdx = -1;
			
			WidgetsStart = wIdxStart;
			
			// We'll play audio through one pin with highest priority (headphones)
			for(int w = wIdxStart; w < wIdxEnd; w++) {
				struct Widget *wstruc = &widgets[WidgetCount];
				
				wstruc->cap = verb(TheCodec, w, 0xF00, 9);
				
				if(WIDGET_TYPE(wstruc) == 4) { //pin
					uint32_t confdef = verb(TheCodec, w, 0xF1C, 0);
					
					uint8_t dev = (confdef >> 20) & 15;
					
					int priority;
					if(dev == 2) {
						// headphone
						priority = 3;
					} else if(dev == 1) {
						// speaker
						priority = 2;
					} else if(dev == 0) {
						// line out
						priority = 1;
						outPinWIdx = WidgetCount;
					}
					
					if(priority > outPinWPriority) {
						outPinWPriority = priority;
						outPinWIdx = WidgetCount;
					}
				}
				
				wstruc->connectionCount = verb(TheCodec, w, 0xF00, 14);
				
				WidgetCount++;
				
				// power on
				verb(TheCodec, w, 0x705, 0);
			}
			
			ThePin = outPinWIdx;
			
			break;
		}
	}
	
	// Hardcoded path
	unmute(0x11, 0, 0, 0);
	unmute(0x11, 1, 0, 0);
	verb(TheCodec, 0x11, 0x70C, 2); //eapd on
	verb(TheCodec, 0x11, 0x707, (1 << 6) | (1 << 7)); // pin out
	
	unmute(0x07, 0, 0, 15);
	unmute(0x07, 1, 0, 15);
	
	unmute(0x22, 1, 0, 0);
	verb(TheCodec, 0x22, 0x701, 0);
	
	unmute(0x03, 1, 0, 0);
	verb16(TheCodec, 0x03, 0x2, STANDARD_48KHZ_16BPS);
	verb(TheCodec, 0x03, 0x706, (1 << 4) | (0 << 0));
	
	HDA[13] = 0xFFFFFFFF;
	HDA[14] = 0xFFFFFFFF;
	
	sys_sleep(10);
	
	memory_barrier();
	outStream[0] |= STREAM_RUN_BIT;
	
	while((outStream[0] & (1 << 29)) == 0) memory_barrier();
	
	sys_sleep(10);
	
	HDA[13] = 0;
	HDA[14] = 0;
	
	// Create canal
	uint32_t canalID = sys_canal_create('AUD\0', 'HDA0');
	uint8_t *AudioLink = (void*) -1;
	do {
		AudioLink = sys_canal_accept(canalID);
	} while(AudioLink == (void*) -1);
	
	sys_signal_send(AudioLink, 0);
	
	while(1) {
		sys_signal_wait(AudioLink, NULL);
		memory_barrier();
		
		while((outStream[0] & (1 << 26)) == 0);
		outStream[0] |= 1 << 26;
		
		//int buf = (outStream[1] / 8192) ^ 1;
		static int buf = 0;
		
		std_copy(audbuf + buf * 4096, AudioLink, 8192);
		buf ^= 1;
		
		memory_barrier();
		sys_signal_send(AudioLink, 0);
	}
	
	while(1) sys_signal_wait(SYS_SIGNAL_WAIT_ANY, NULL);
}
