#ifndef LUMA_STD_H
#define LUMA_STD_H

#include<stddef.h>
#include<stdint.h>

#if defined(__has_builtin) && __has_builtin(__builtin_memcpy)
#define std_copy __builtin_memcpy
#else
static inline void std_copy(void *dst, const void *src, size_t len) {
	char *d = dst;
	const char *s = src;
	while(len--) {
		*(d++) = *(s++);
	}
}
#endif

static inline void std_move(void *dst, const void *src, size_t len) {
	if(src > dst) {
		char *d = dst;
		const char *s = src;
		while(len--) {
			*(d++) = *(s++);
		}
	} else {
		char *d = dst + len;
		const char *s = src + len;
		while(len--) {
			*(--d) = *(--s);
		}
	}
}

static inline void std_w8(uint8_t *dst, uint8_t b, size_t len) {
	while(len--) {
		*dst = b;
		dst++;
	}
}

typedef struct {
	uint16_t len;
	uint8_t data[];
} Str16;
#define WALK_STR16(x) ((Str16*)((uintptr_t)(x)+sizeof(Str16)+((Str16*)(x))->len))

typedef struct {
	uint16_t cap;
	uint16_t len;
	uint8_t data[];
} DynStr16;
#define DYN16_TO_ST(x) ((Str16*)((uintptr_t) x + offsetof(DynStr16, len)))

static inline char* std_bytefind(size_t len, const char *data, const char what) {
	while(len--) {
		if(*data == what) {
			return (char*) data;
		}
		data++;
	}
	return NULL;
}

static inline unsigned char std_bytecomp(size_t len, const void *one_, const void *two_) {
	const char *one = one_, *two = two_;
	while(len--) {
		unsigned char asdf = *(one++) - *(two++);
		if(asdf) return asdf;
	}
	return 0;
}

#define std_min(a, b) ({__typeof__(a) _a = (a); __typeof__(b) _b = (b); _a<_b?_a:_b;})
#define std_max(a, b) ({__typeof__(a) _a = (a); __typeof__(b) _b = (b); _a>_b?_a:_b;})

//#define S16_ARG(str) ((Str16*) &(struct {uint16_t len; char content[sizeof(str) - 1]}) {sizeof(str) - 1, (str)})

#endif
