#ifndef _LUMA_SYS_H
#define _LUMA_SYS_H

#include<stdint.h>
#include<stddef.h>

#define memory_barrier() asm volatile("" : : : "memory")
#define STRINGIFY2(v) #v
#define STRINGIFY(v) STRINGIFY2(v)

static __attribute__((always_inline)) inline uint32_t syscall1(uint32_t eax, uint32_t edi, uint32_t esi, uint32_t edx) {
	uint32_t ret;
	asm volatile("int $0xEC" : "=a"(ret) : "a"(eax), "S"(esi), "D"(edi), "d"(edx));
	return ret;
}

static __attribute__((always_inline)) inline uint32_t sys5call1(uint32_t eax, uint32_t edi, uint32_t esi, uint32_t edx, uint32_t ecx, uint32_t ebx) {
	uint32_t ret;
	asm volatile("int $0xEC" : "=a"(ret) : "a"(eax), "S"(esi), "D"(edi), "d"(edx), "c"(ecx), "b"(ebx));
	return ret;
}

#define SYS_MAP_PHYS_ANY ((void*)-1)
#define SYS_MAP_PHYS_CONTIGUOUS ((void*)-2)
#define SYS_MAP_VIRT_ANY ((void*)-1)
static __attribute__((always_inline)) inline void *sys_vpm_map(void *virt, void *phys, size_t length) {
	return (void*) syscall1(0, (uint32_t) virt, (uint32_t) phys, length);
}

static __attribute__((always_inline)) inline uint32_t sys_canal_create(uint32_t nameLow, uint32_t nameHigh) {
	return syscall1(1, nameLow, nameHigh, 0);
}

static __attribute__((always_inline)) inline void *sys_link_create(uint32_t nameLow, uint32_t nameHigh, size_t areaSize) {
	return (void*) syscall1(2, nameLow, nameHigh, areaSize);
}

#define SYS_CANAL_ACCEPT_ANY ((uint32_t)-1)
static __attribute__((always_inline)) inline void *sys_canal_accept(uint32_t canalId) {
	return (void*) syscall1(3, canalId, 0, 0);
}

static __attribute__((always_inline)) inline void sys_signal_send(void *area, uint32_t meta) {
	syscall1(4, (uint32_t) area, meta, 0);
}

#define SYS_SIGNAL_WAIT_ANY ((void*)-1)
static __attribute__((always_inline)) inline void *sys_signal_wait(void *area, uint32_t *meta) {
	uint32_t eax, edx;
	asm volatile("int $0xEC" : "=a"(eax), "=d"(edx) : "a"(5), "D"(area) : "ebx");
	if(meta) *meta = edx;
	return (void*) eax;
}
static __attribute__((always_inline)) inline void *sys_signal_wait3(void *area, uint32_t *meta, uint32_t *canalId) {
	uint32_t eax, edx, ebx;
	asm volatile("int $0xEC" : "=a"(eax), "=d"(edx), "=b"(ebx) : "a"(5), "D"(area));
	if(meta) *meta = edx;
	if(canalId) *canalId = ebx;
	return (void*) eax;
}

static __attribute__((always_inline)) inline uint32_t sys_pci_claim(uint32_t cscp) {
	return syscall1(6, cscp, 0, 0);
}

static __attribute__((always_inline)) inline uint32_t sys_pci_read(uint32_t id, uint8_t offset) {
	return syscall1(7, id, offset, 0);
}

static __attribute__((always_inline)) inline void sys_vpm_unmap(void *virt, size_t length) {
	syscall1(8, (uint32_t) virt, length, 0);
}

static __attribute__((always_inline)) inline void *sys_vpm_get_phys(void *virt) {
	return (void*) syscall1(9, (uint32_t) virt, 0, 0);
}

static __attribute__((always_inline)) inline void sys_sleep(uint32_t millis) {
	syscall1(10, millis, 0, 0);
}

static __attribute__((always_inline)) inline void sys_pci_write_1(uint32_t id, uint8_t offset, uint8_t val) {
	syscall1(11, id, offset, val);
}

static __attribute__((always_inline)) inline void sys_pci_write_2(uint32_t id, uint8_t offset, uint16_t val) {
	syscall1(12, id, offset, val);
}

static __attribute__((always_inline)) inline void sys_pci_write_4(uint32_t id, uint8_t offset, uint32_t val) {
	syscall1(13, id, offset, val);
}

static __attribute__((always_inline)) inline void sys_irq_claim(uint32_t irq, void(*handler)()) {
	syscall1(14, irq, (uintptr_t) handler, 0);
}

static __attribute__((always_inline)) inline void sys_wait_for_irq(uint32_t i) {
	syscall1(15, i, 0, 0);
}

static __attribute__((always_inline)) inline void sys_create_thread(void *cod) {
	syscall1(16, (uintptr_t) cod, 0, 0);
}

static __attribute__((always_inline)) inline size_t sys_spawn() {
	return syscall1(17, 0, 0, 0);
}

static __attribute__((always_inline)) inline uintptr_t sys_abuse_map(size_t handle, void **virtMy, void **virtTheir, void *phys, size_t length) {
	memory_barrier();
	return sys5call1(18, handle, (uintptr_t) virtMy, (uintptr_t) virtTheir, (uintptr_t) phys, length);
}

static __attribute__((always_inline)) inline int sys_release(size_t handle) {
	return syscall1(19, handle, 0, 0);
}

#define SYS_STATE_X86_EIP 0
static __attribute__((always_inline)) inline int sys_abuse_state(size_t handle, uintmax_t type, uintmax_t val) {
	return syscall1(20, handle, type, (uintptr_t) val);
}

static __attribute__((always_inline)) inline int sys_exit() {
	return syscall1(21, 0, 0, 0);
}

#define KEV_TYPE_NONE 0
#define KEV_TYPE_LINK_REQUEST 1
#define KEV_TYPE_LINK_SIGNAL 2
#define KEV_TYPE_LINK_CLOSE 3
typedef struct __attribute__((packed)) {
	uint8_t type;
	uint32_t canalId;
	void *area;
	uint32_t meta;
} KEvent;

static __attribute__((always_inline)) inline int sys_event_wait(KEvent *kev) {
	int ret = syscall1(22, (uintptr_t) (void*) kev, 0, 0);
	memory_barrier();
	return ret;
}

static __attribute__((always_inline)) inline uint32_t ind(uint32_t p) {
	uint32_t ret;
	asm volatile("inl %%dx, %%eax" : "=a"(ret) : "d"(p));
	return ret;
}

static __attribute__((always_inline)) inline uint16_t inw(uint32_t p) {
	uint16_t ret;
	asm volatile("inw %%dx, %%ax" : "=a"(ret) : "d"(p));
	return ret;
}

static __attribute__((always_inline)) inline uint8_t inb(uint32_t p) {
	uint8_t ret;
	asm volatile("inb %%dx, %%al" : "=a"(ret) : "d"(p));
	return ret;
}

static __attribute__((always_inline)) inline void outb(uint32_t p, uint8_t v) {
	asm volatile("outb %%al, %%dx" : : "d"(p), "a"(v));
}

static __attribute__((always_inline)) inline void outw(uint32_t p, uint16_t v) {
	asm volatile("outw %%ax, %%dx" : : "d"(p), "a"(v));
}

static __attribute__((always_inline)) inline void outd(uint32_t p, uint32_t v) {
	asm volatile("outl %%eax, %%dx" : : "d"(p), "a"(v));
}

#if __BYTE_ORDER == __LITTLE_ENDIAN
#define BIG_TO_HOST_ENDIAN32(x) __builtin_bswap32(x)
#define BIG_TO_HOST_ENDIAN16(x) __builtin_bswap16(x)
#elif __BYTE_ORDER == __BIG_ENDIAN
#define BIG_TO_HOST_ENDIAN32(x) (x)
#define BIG_TO_HOST_ENDIAN16(x) (x)
#else
#error "PDP??"
#endif

#endif
