#include"../sys.h"

// Currently, the driver waits for any requests and performs them
// This method fails for active USB devices (rather than passive)
//
// Should either poll for signals or switch to multithreading, once
// it becomes a need.

asm("modentry:\n"
"mov $stack + 4096, %esp\n"
"call startc");

#include"host.h"

#define TD_ACTIVE (1 << 23)
#define TD_MAXERRS(x) ((x) << 27)
#define TD_TOGGLE(x) ((!!(x)) << 19)
#define TD_MAXLEN(sz) (((sz) - 1) << 21)
#define TD_MAXLEN_0 (0x7FF << 21)
#define TD_DEVADDR(i) ((i) << 8)
#define TD_DEVENDP(i) ((i) << 15)
#define TD_LSD (1 << 26)
#define TD_SPD (1 << 29)
#define TD_IOC (1 << 24)
#define TD_DEPTH_FIRST (4)

char __attribute__((aligned(16))) stack[4096];

size_t hcCount = 0;
HC __attribute__((aligned(16))) hcs[4];

size_t deviceCount = 0;
USBDevice *devices;

// FOR DEBUG PURPOSES {
static volatile uint8_t *Framebuffer;
uint32_t asdfasdf = 28;
void dbg(int j, uint8_t b) {
	for(int i = 0; i < 8; i++) {
		Framebuffer[j * 9 + 10 - i] = (b & (1 << i)) ? 85 : 40;
	}
	Framebuffer[j * 9 + 11] = 0;
}
// }

int HostSendBulk(USBDevice *dev, size_t maxPacketSize, int *toggle, int endpoint, void *buf, size_t len) {
	size_t sz = len;
	int num = (sz + maxPacketSize - 1) / maxPacketSize;
	
	TD tds[num] __attribute__((aligned(16)));
	
	for(int i = 0;; i++, sz -= maxPacketSize) {
		tds[i] = (TD) {
			.link = (i == num - 1 ? 1 : ((uintptr_t) sys_vpm_get_phys(&tds[i + 1]))) | TD_DEPTH_FIRST,
			.control = TD_ACTIVE | TD_MAXERRS(3) | ((endpoint & 128) ? TD_SPD : 0) | (i == num - 1 ? TD_IOC : 0),
			.token = TD_MAXLEN(sz > maxPacketSize ? maxPacketSize : sz) | TD_TOGGLE(*toggle) | TD_DEVADDR(dev->id) | TD_DEVENDP(endpoint & 0x0F) | ((endpoint & 128) ? IN : OUT),
			.buffer = (uintptr_t) sys_vpm_get_phys(&buf[i * maxPacketSize])
		};
		*toggle ^= 1;
		if(sz <= maxPacketSize) break;
	}
	
	while(inw(hcs[dev->hc].io + 2) & 3) {
		outw(hcs[dev->hc].io + 2, 3); // clear usbint
	}
	
	QH *masterQ = &hcs[dev->hc].masterQueue;
	
	QH qh __attribute__((aligned(16)));
	qh.link = 1;
	
	memory_barrier();
	
	qh.element = (uint32_t) sys_vpm_get_phys(&tds[0]);
	masterQ->element = ((uint32_t) sys_vpm_get_phys(&qh)) | 2;
	
	// Wait until complete.
	// Polling doesn't seem to work reliably. Sometimes USBINT just isn't set.
	// From iPXE source, though can't find a relation:
	/* UHCI defers interrupts (including short packet detection)
	* until the end of the frame.  This can result in bulk IN
	* endpoints remaining halted for much of the time, waiting
	* for software action to reset the data toggles.  We
	* therefore ignore USBSTS and unconditionally poll all
	* endpoints for completed transfer descriptors.
	*
	* As with EHCI, we trust that completion handlers are minimal
	* and will not do anything that could plausibly affect the
	* endpoint list itself.
	*/
	for(int i = 0; i < num; i++) {
		while(tds[i].control & TD_ACTIVE) memory_barrier();
		
		if(tds[i].control & 0x00FF0000) {
			return 0; //fail
		}
	}
	
	return 1;
}

/* buf must be physically contiguous */
int HostSendRequest(USBDevice *dev, void *buf, USBStandardRequest req) {
	size_t sz = req.length;
	int num = 2 + (sz + dev->maxPacketSize - 1) / dev->maxPacketSize;
	
	TD tds[num] __attribute__((aligned(16)));
	
	tds[0] = (TD) {
		.link = (sz == 0 ? 1 : (uintptr_t) sys_vpm_get_phys(&tds[1])) | TD_DEPTH_FIRST,
		.control = TD_ACTIVE | TD_MAXERRS(3) | (sz == 0 ? TD_IOC : 0),
		.token = TD_MAXLEN(8) | TD_TOGGLE(0) | TD_DEVADDR(dev->id) | SETUP,
		.buffer = (uintptr_t) sys_vpm_get_phys(&req)
	};
	
	for(int i = 1;; i++, sz -= dev->maxPacketSize) {
		tds[i] = (TD) {
			.link = (i == num - 2 ? 1 : ((uintptr_t) sys_vpm_get_phys(&tds[i + 1]))) | TD_DEPTH_FIRST,
			.control = TD_ACTIVE | TD_MAXERRS(3) | ((req.requestType & 128) ? TD_SPD : 0) | (i == num - 2 ? TD_IOC : 0),
			.token = TD_MAXLEN(sz > dev->maxPacketSize ? dev->maxPacketSize : sz) | TD_TOGGLE(i & 1) | TD_DEVADDR(dev->id) | ((req.requestType & 128) ? IN : OUT),
			.buffer = (uintptr_t) sys_vpm_get_phys(&buf[(i - 1) * dev->maxPacketSize])
		};
		if(sz <= dev->maxPacketSize) break;
	}
	
	tds[num - 1] = (TD) {
		.link = 1,
		.control = TD_ACTIVE | TD_MAXERRS(3),
		.token = TD_MAXLEN_0 | TD_TOGGLE(1) | TD_DEVADDR(dev->id) | ((req.requestType & 128) ? OUT : IN),
		.buffer = 0
	};
	
	while(inw(hcs[dev->hc].io + 2) & 1) {
		outw(hcs[dev->hc].io + 2, 3); // clear usbint
	}
	
	QH *masterQ = &hcs[dev->hc].masterQueue;
	
	QH qh __attribute__((aligned(16)));
	qh.link = 1;
	
	memory_barrier();
	
	qh.element = (uint32_t) sys_vpm_get_phys(&tds[0]);
	masterQ->element = ((uint32_t) sys_vpm_get_phys(&qh)) | 2;
	
	sys_sleep(15);
	
	// Wait for usbint
	memory_barrier();
	while((inw(hcs[dev->hc].io + 2) & 1) == 0);
	
	outw(hcs[dev->hc].io + 2, 3); // clear usbint
	
	qh.element = (uint32_t) sys_vpm_get_phys(&tds[num - 1]);
	masterQ->element = ((uint32_t) sys_vpm_get_phys(&qh)) | 2;
	
	// Wait until status packet becomes inactive
	do {
		memory_barrier();
	} while(tds[num - 1].control & TD_ACTIVE);
	
	// Return whether success
	return (tds[num - 1].control & (0xFF0000)) == 0;
}

static int HCsInit() {
	// Go through all EHCIs and disable them
	while(1) {
		uint32_t hcId = sys_pci_claim(0x0C032000);
		
		if(hcId == -1) {
			break;
		}
		
		volatile uint32_t *usbbase = sys_vpm_map(SYS_MAP_VIRT_ANY, (void*) (sys_pci_read(hcId, 16) & ~15), 4096);
		volatile uint32_t *opregs = (uint32_t*) ((uintptr_t) usbbase + (usbbase[0] & 0xFF));
		
		uint8_t caps = (usbbase[2] & 0xFF00) >> 8;
		if(caps < 0x40) {
			return 0;
		} else {
			sys_pci_write_1(hcId, caps + 3, 1);
			while((sys_pci_read(hcId, caps) & (1 << 16)) != 0);
			while((sys_pci_read(hcId, caps) & (1 << 24)) == 0);
			sys_pci_write_4(hcId, caps + 4, 0);
		}
		
		opregs[0] &= ~1;
		
		do {
			memory_barrier();
		} while((opregs[1] & 4096) == 0);
		
		opregs[0] |= 2;
		
		do {
			memory_barrier();
		} while(opregs[0] & 2);
		
		opregs[1] = opregs[1];
		opregs[2] = 0;
		opregs[4] = 0;
		opregs[0] = 0x80000;
		opregs[16] = 0;
		
		sys_vpm_unmap(usbbase, 4096);
	}
	
	int success = 1;
	
	// Go through all UHCIs
	while(1) {
		uint32_t hcId = sys_pci_claim(0x0C030000);
		
		if(hcId == -1) {
			break;
		}
		
		HC *hc = &hcs[hcCount];
		hc->io = sys_pci_read(hcId, 0x20) & ~3;
		hc->frameList = sys_vpm_map(SYS_MAP_VIRT_ANY, SYS_MAP_PHYS_ANY, 4096);
		hc->masterQueue = (QH) {1, 1};
		for(int i = 0; i < 1024; i++) {
			hc->frameList[i] = ((uint32_t) sys_vpm_get_phys(&hc->masterQueue)) | 2;
		}
		
		sys_pci_write_2(hcId, 0xC0, 0x8F00);
		
		outw(hc->io + 0, 2);
		while(inw(hc->io + 2) & 2);
		
		outw(hc->io + 4, 0);
		outw(hc->io + 0, 0);
		outw(hc->io + 2, 0xFFFF);
		outb(hc->io + 12, 64);
		outd(hc->io + 8, (uint32_t) sys_vpm_get_phys(hc->frameList));
		outw(hc->io + 6, 0);
		outw(hc->io + 0, 1 | 128);
		
		for(int p = 0; p < 2; p++) {
			outw(hc->io + 16 + p+p, 512);
			sys_sleep(80);
			outw(hc->io + 16 + p+p, 0);
			
			for(int try = 0; try < 10; try++) {
				sys_sleep(10);
				uint16_t sc = inw(hc->io + 16 + p+p);
				//if((sc & 1) == 0) break;
				if(sc & (2 | 8)) outw(hc->io + 16 + p+p, sc);
				if(sc & 4) break;
				outw(hc->io + 16 + p+p, sc | 4);
			}
			
			sys_sleep(100);
			
			if(inw(hc->io + 16 + p+p) & 1) {
				dbg(asdfasdf++, 2 * hcCount + p + 128);
				
				USBDevice *dev = &devices[deviceCount];
				dev->hc = hcCount;
				dev->id = deviceCount + 1;
				dev->maxPacketSize = 8;
				if(USBDeviceDiscover(dev) == 0) { /* The rest of the fields are filled here */
					success = 0;
				}
				deviceCount++;
			}
		}
		
		hcCount++;
	}
	
	return success;
}

void startc() {
	devices = sys_vpm_map(SYS_MAP_VIRT_ANY, SYS_MAP_PHYS_ANY, 4096);
	
	Framebuffer = sys_vpm_map(SYS_MAP_VIRT_ANY, (void*) 0xA0000, 32000);
	
	if(HCsInit()) {
		// Let other drivers know we exist
		sys_canal_create('USB\0', 'UHCI');
		
		// Main loop
		while(1) {
			sys_canal_accept(SYS_CANAL_ACCEPT_ANY);
			
			uint32_t meta, canal;
			void* area = sys_signal_wait3(SYS_SIGNAL_WAIT_ANY, &meta, &canal);
			for(int i = 0; i < deviceCount; i++) {
				if(devices[i].canal == canal) {
					if(devices[i].respond) {
						devices[i].respond(&devices[i], area, meta);
					}
					break;
				}
			}
		}
	} else {
		while(1) {
			sys_signal_wait(SYS_SIGNAL_WAIT_ANY, NULL);
		}
	}
}
