#pragma once

#include<stdint.h>

typedef enum {
	SETUP = 0x2D, IN = 0x69, OUT = 0xE1
} PacketType;

typedef struct USBDevice {
	uint8_t hc;
	uint8_t addr;
	uint8_t speedy;
	uint8_t id;
	uint8_t maxPacketSize;
	uint32_t canal;
	union {
		struct {
			int endpIn;
			int endpOut;
			int maxPacketSize;
			int toggleIn;
			int toggleOut;
			uint16_t sectorSize;
		} bbb;
	};
	int(*respond)(struct USBDevice *dev, void *area, uint32_t meta);
} USBDevice;

typedef struct __attribute((packed)) {
	uint8_t requestType;
	uint8_t request;
	uint16_t value;
	uint16_t index;
	uint16_t length;
} USBStandardRequest;

typedef struct {
	uint8_t sz;
	uint8_t type;
	uint16_t usbVer;
	uint8_t clazz;
	uint8_t subclazz;
	uint8_t protocol;
	uint8_t maxPacketSize;
	uint16_t idVendor;
	uint16_t idProduct;
	uint16_t devVer;
	uint8_t stringManufacturer;
	uint8_t stringProduct;
	uint8_t stringSerial;
	uint8_t configurationCount;
} USBDeviceDescriptor;

typedef struct {
	uint8_t sz;
	uint8_t type;
	uint16_t szTotal;
	uint8_t interfaceCount;
	uint8_t id;
	uint8_t string;
	uint8_t attributes;
	uint8_t maxPower;
} USBConfigurationDescriptor;

typedef struct {
	uint8_t sz;
	uint8_t type;
	uint8_t interface;
	uint8_t alternate;
	uint8_t endpointCount;
	uint8_t clazz;
	uint8_t subclazz;
	uint8_t protocol;
	uint8_t string;
} USBInterfaceDescriptor;

typedef struct {
	uint8_t sz;
	uint8_t type;
	uint8_t endpoint;
	uint8_t attributes;
	uint16_t maxPacketSize;
	uint8_t interval;
} USBEndpointDescriptor;

int USBDeviceDiscover(USBDevice*);