#include<stddef.h>
#include<stdint.h>

#include"ebml.h"

#include"reader.h"

void ebml_reader_init(EBMLReader *this) {
	EEBIE_MEMSET(this, 0, sizeof(*this));
	this->state = EBMLRS_WAITING_FOR_ELEMENT_ID;
	this->currentDepth = 0;
}

static uint64_t VARINT_MASKS[] = {0, 0x80L, 0xC000L, 0xE00000L, 0xF0000000L, 0xF800000000L, 0xFC0000000000L, 0xFE000000000000L, 0L};

#include"../../std.h"
static inline void pr(int c) {
	asm volatile("outb %%al, $0xE9" :: "a"(c) :);
}

static inline void pri(size_t num) {
	char buf[25] = {};
	int i = 16;
	
	do {
		buf[--i] = num % 10;
		num /= 10;
	} while(num);
	
	for(; i < 16; i++) {
		pr(buf[i] + '0');
	}
}

static inline void prs(const Str16 *str) {
	for(int i = 0; i < str->len; i++) {
		pr(str->data[i]);
	}
}

static inline void prcs(const char *str) {
	while(*str) {
		pr(*str);
		str++;
	}
}

#define I(x) ((uint64_t)(x))
static int read_varint(const uint8_t *data, size_t length, uint64_t *result) {
	if(data[0] & 0x80) {
		*result = data[0];
		
		return 1;
	} else if(data[0] & 0xC0) {
		if(length < 2) {
			return 0;
		}
		
		*result = (I(data[0]) << 8) | data[1];
		
		return 2;
	} else if(data[0] & 0xE0) {
		if(length < 3) {
			return 0;
		}
		
		*result = (I(data[0]) << 16) | (I(data[1]) << 8) | data[2];
		
		return 3;
	} else if(data[0] & 0xF0) {
		if(length < 4) {
			return 0;
		}
		
		*result = (I(data[0]) << 24) | (I(data[1]) << 16) | (I(data[2]) << 8) | data[3];
		
		return 4;
	} else if(data[0] & 0xF8) {
		if(length < 5) {
			return 0;
		}
		
		*result = (I(data[0]) << 32) | (I(data[1]) << 24) | (I(data[2]) << 16) | (I(data[3]) << 8) | data[4];
		
		return 5;
	} else if(data[0] & 0xFC) {
		if(length < 6) {
			return 0;
		}
		
		*result = (I(data[0]) << 40) | (I(data[1]) << 32) | (I(data[2]) << 24) | (I(data[3]) << 16) | (I(data[4]) << 8) | data[5];
		
		return 6;
	} else if(data[0] & 0xFE) {
		if(length < 7) {
			return 0;
		}
		
		*result = (I(data[0]) << 48) | (I(data[1]) << 40) | (I(data[2]) << 32) | (I(data[3]) << 24) | (I(data[4]) << 16) | (I(data[5]) << 8) | data[6];
		
		return 7;
	} else if(data[0] == 0x01) {
		if(length < 8) {
			return 0;
		}
		
		*result = (I(data[1]) << 48) | (I(data[2]) << 40) | (I(data[3]) << 32) | (I(data[4]) << 24) | (I(data[5]) << 16) | (I(data[6]) << 8) | data[7];
		
		return 8;
	}
	
	return -1;
}

static int get_varint(const uint8_t *data, size_t length, uint64_t *result) {
	int ret = read_varint(data, length, result);
	
	if(ret >= 0) {
		*result &= ~VARINT_MASKS[ret];
	}
	
	return ret;
}

int ebml_reader_feed(EBMLReader *this, const uint8_t *data, size_t length) {
	if(length == 0) {
		return 0;
	}
	
	size_t eaten = -1;
	
	if(this->state == EBMLRS_WAITING_FOR_ELEMENT_ID) {
		
		uint64_t elId;
		
		int status = read_varint(data, length, &elId);
		
		if(status <= 0) {
			return status;
		}
		
		this->state = EBMLRS_WAITING_FOR_ELEMENT_LENGTH;
		this->inside.id = elId;
		
		eaten = status;
		
	} else if(this->state == EBMLRS_WAITING_FOR_ELEMENT_LENGTH) {
		
		uint64_t elLength;
		
		int status = get_varint(data, length, &elLength);
		
		if(status <= 0) {
			return status;
		}
		
		this->inside.type = this->eventEnterElement(this, this->inside.id, elLength);
		
		if(this->inside.type == EBML_TREE) {
			this->state = EBMLRS_WAITING_FOR_ELEMENT_ID;
		} else {
			this->state = EBMLRS_WAITING_FOR_ELEMENT_DATA;
			
			this->inside.length = elLength;
		}
		
		this->idStack[this->currentDepth] = this->inside.id;
		this->stack[this->currentDepth] = elLength + status;
		
		this->currentDepth++;
		
		eaten = status;
		
	} else if(this->state == EBMLRS_WAITING_FOR_ELEMENT_DATA) {
	
		size_t realLen = this->stack[this->currentDepth - 1] < length ? this->stack[this->currentDepth - 1] : length;
		
		if(this->eventDataChunk) {
			this->eventDataChunk(this, data, realLen);
		}
		
		eaten = realLen;
		
	}
	
	for(int i = 0; i < this->currentDepth; i++) {
		this->stack[i] -= eaten;
	}
	
	while(this->currentDepth > 0 && this->stack[this->currentDepth - 1] == 0) {
		
		if(this->eventExitElement) {
			this->eventExitElement(this);
		}
		
		this->currentDepth--;
		
		this->state = EBMLRS_WAITING_FOR_ELEMENT_ID;
		
	}
	
	return eaten;
}
