#include"fileswindow.h"

#include"../std.h"
#include"../sys.h"
#include"../fs/fs.h"
#include"wm.h"
#include"tiny.h"

#define dynstr16_ends_with(ds16, cstr) (((ds16)->len < sizeof(cstr) - 1) ? 0 : !std_bytecomp(sizeof(cstr) - 1, (cstr), (ds16)->data + (ds16)->len - sizeof(cstr) + 1))

struct Ctx {
	Str16 *cd;
	size_t cdSize;
} ;

void __attribute__((naked)) audio_thread() {
	asm(
		"movl %eax, %ebp\n"
		"movl $0, %eax\n"
		"movl $-1, %edi\n"
		"movl $-1, %esi\n"
		"movl $4096, %edx\n"
		"int $0xEC\n"
		"leal 4096(%eax), %esp\n"
		"pushl %ebp\n"
		"call caudio_thread\n"
	);
}
void caudio_thread(void*) {
	void *FSLink = sys_link_create('FS\0\0', 'ROOT', 4096);
	
	LumaFSCommand *cmd = FSLink;
	cmd->op = LUMA_FS_COMMAND_READ;
	cmd->read.offset = 44;
	cmd->read.bytes = 1024 * 1024;
	Str16 *path = (Str16*) cmd->read.path;
	std_copy(path->data, "duane.wav", path->len = 9);
	WALK_STR16(path)->len = 0; //terminator
	
	uint8_t *wavebuffer = sys_vpm_map(SYS_MAP_VIRT_ANY, SYS_MAP_PHYS_ANY, 1024 * 1024);
	size_t wavebufferoffset = 0;
	while(1) {
		memory_barrier();
		sys_signal_send(FSLink, 1);
		
		uint32_t dasize;
		sys_signal_wait(FSLink, &dasize);
		
		std_copy(wavebuffer + wavebufferoffset, FSLink, dasize);
		wavebufferoffset += dasize;
		
		if(dasize != 4096) break;
	}
	
	wavebufferoffset = 0;
	
	uint8_t *AudLink = sys_link_create('AUD\0', 'HDA0', 8192);
	
	while(1) {
		if(wavebufferoffset == 1024 * 1024) wavebufferoffset = 0;
		
		sys_signal_wait(AudLink, NULL);
		std_copy(AudLink, wavebuffer + wavebufferoffset, 8192);
		wavebufferoffset += 8192;
		sys_signal_send(AudLink, 0);
	}
}

static void update_listing(struct GWindow *w) {
	struct Ctx *ctx = w->ud;
	
	struct GList *list = (struct GList*) w->child;
	
	glist_clear(list);
	
	LumaFSCommand *cmd = FSLink;
	cmd->op = LUMA_FS_COMMAND_TRAVERSE;
	
	std_copy(cmd->traverse.path, ctx->cd, ctx->cdSize);
	
	sys_signal_send(FSLink, 0);
	sys_signal_wait(FSLink, NULL);
	
	for(LumaFSDirectoryEntry *e = FSLink;; e = (LumaFSDirectoryEntry*) ((uintptr_t) e + sizeof(*e) + e->name.len)) {
		glist_add(list, g_dynstr16_new(e->name.len, e->name.len, e->name.data), (void*) (uintptr_t) e->type);
		
		if(e->type & LUMA_FS_DIR_ENTRY_LAST) {
			break;
		}
	}
}

static int files_window_view_cmd(void *ud, const DynStr16 *cmdline) {
	if(!is_free_space()) return 0;
	
	struct GWindow *w = ud;
	
	struct GList *list = (struct GList*) w->child;
	
	struct Ctx *ctx = w->ud;
	
	if(list->cursor >= list->childCount) return 0;
	
	if(((uintptr_t) (list->children[list->cursor].ud)) & LUMA_FS_DIR_ENTRY_ISDIR) {
		Str16 *newud = tiny_realloc(ctx->cd, ctx->cdSize + sizeof(*newud) + list->children[list->cursor].text->len);
		
		if(newud) {
			ctx->cd = newud;
			
			Str16 *end = (void*) ((uintptr_t) ctx->cd + ctx->cdSize - sizeof(*end));
			
			end->len = list->children[list->cursor].text->len;
			std_copy(end->data, list->children[list->cursor].text->data, end->len);
			
			// Terminator.
			WALK_STR16(end)->len = 0;
			
			ctx->cdSize += sizeof(*newud) + list->children[list->cursor].text->len;
			
			update_listing(w);
		} else {
			// TODO: log
		}
	} else if(dynstr16_ends_with(list->children[list->cursor].text, ".txt")) {
		size_t amount = 100;
		
		Str16 *newterm = DYN16_TO_ST(list->children[list->cursor].text);
		
		LumaFSCommand *cmd = FSLink;
		cmd->op = LUMA_FS_COMMAND_READ;
		cmd->read.offset = 0;
		cmd->read.bytes = amount;
		std_copy(cmd->read.path, ctx->cd, ctx->cdSize - sizeof(*newterm));
		std_copy(cmd->read.path + ctx->cdSize - sizeof(*newterm), newterm, sizeof(*newterm) + newterm->len);
		WALK_STR16(cmd->read.path + ctx->cdSize - sizeof(*newterm))->len = 0; //terminator

		memory_barrier();
		sys_signal_send(FSLink, 0);
		
		uint32_t dasize;
		sys_signal_wait(FSLink, &dasize);
		
		memory_barrier();
		
		DynStr16 *str = g_dynstr16_new(512, dasize, FSLink);
		
		struct GText *bigtxt = gtext_new(0, 0, 0, 0, NULL, str);
		
		static const char name[] = "Viewer";
		struct GWindow *viewwnd = gwindow_new(0, 0, 0, 0, g_dynstr16_new(sizeof(name) - 1, sizeof(name) - 1, name), (struct GObj*) bigtxt, 0);
		
		add_window(viewwnd);
		
		return 1;
	} else if(dynstr16_ends_with(list->children[list->cursor].text, ".wav")) {
		static const char name[] = "Audio Player";
		struct GWindow *playerwnd = gwindow_new(0, 0, 0, 0, g_dynstr16_new(sizeof(name) - 1, sizeof(name) - 1, name), NULL, 0);
		
		add_window(playerwnd);
		
		sys_create_thread(audio_thread);
		
		return 1;
	}
	
	return 0;
}

static int files_window_back_cmd(void *ud, const DynStr16 *cmdline) {
	struct GWindow *w = ud;
	
	struct Ctx *ctx = w->ud;
	
	Str16 *pathPrev = NULL;
	Str16 *path = ctx->cd;
	
	if(path->len == 0) {
		return 0;
	}
	
	while(path->len) {
		pathPrev = path;
		path = WALK_STR16(path);
	}
	
	// Turn term into terminator.
	ctx->cdSize -= pathPrev->len;
	pathPrev->len = 0;
	
	update_listing(w);
	
	return 1;
}

void FilesWindowCreate() {
	if(!is_free_space()) return;
	
	struct GList *list = glist_new(0, 0, 0, 0);
	
	static const char name[] = "File Explorer";
	struct GWindow *w = gwindow_new(0, 0, 0, 0, g_dynstr16_new(sizeof(name) - 1, sizeof(name) - 1, name), (struct GObj*) list, 0);
	
	struct Ctx *ctx = tiny_malloc(sizeof(*ctx));
	
	ctx->cd = tiny_malloc(ctx->cdSize = sizeof(*ctx->cd));
	ctx->cd->len = 0; // root
	
	w->ud = ctx;
	
	update_listing(w);
	
	static const char viewcmd[] = "open";
	gwindow_regcmd(w, g_str16_new(sizeof(viewcmd) - 1, viewcmd), w, files_window_view_cmd);
	
	static const char backcmd[] = "back";
	gwindow_regcmd(w, g_str16_new(sizeof(backcmd) - 1, backcmd), w, files_window_back_cmd);
	
	add_window(w);
}
