#pragma once

#include"../std.h"
#include"../ps2/api.h"

typedef uint16_t Coord;

struct GObj;
struct GContainer;

#define SHIFT_LEFT 1
#define SHIFT_RIGHT 2
#define ALT_LEFT 1
#define ALT_RIGHT 2
#define SYS_LEFT 1
#define SYS_RIGHT 2
struct GKeyEvent {
	uint8_t scancode;
	uint8_t press;
	uint8_t shift;
	uint8_t alt;
	uint8_t sys;
};

typedef void(*RenderFunc)(struct GObj*);
typedef void(*KeyReactFunc)(struct GObj*, struct GKeyEvent);
typedef void(*FitterFunc)(struct GObj*);
typedef void(*DestroyFunc)(struct GObj*);

struct GObj {
	RenderFunc onRender;
	KeyReactFunc onKey;
	DestroyFunc destroy;
	FitterFunc fitter;
	Coord x, y;
	Coord w, h;
	uint8_t focus;
	void *ud;
};

struct GText {
	struct GObj;
	
	DynStr16 *placeholder;
	DynStr16 *data;
};
struct GText *gtext_new(Coord x, Coord y, Coord w, Coord h, DynStr16 *placeholder, DynStr16 *str);

typedef uint8_t GContainerChildMeta;
struct GContainerChild {
	GContainerChildMeta meta;
	struct GObj *obj;
};
struct GContainer {
	struct GObj;
	
	size_t childCount;
	struct GContainerChild *children;
	
	uint8_t meta;
};
struct GContainer *gcont_new(Coord x, Coord y, Coord w, Coord h, FitterFunc fitter);
int gcont_add(struct GContainer*, GContainerChildMeta, struct GObj*);
void gcont_set(struct GContainer*, size_t index, GContainerChildMeta, struct GObj*);
void gcont_del(struct GContainer*, size_t index);

void gcont_splitter_v(struct GObj*);
void gcont_splitter_h(struct GObj*);

typedef int(*GCMDFunc)(void*, const DynStr16 *line);
struct Command {
	Str16 *name;
	void *ud;
	GCMDFunc run;
};
struct CommandList {
	size_t count;
	struct Command *array;
};

struct GWindow {
	struct GObj;
	
	DynStr16 *name;
	
	struct GObj *child;
	
	size_t owner;
	
	struct CommandList cmds;
};
struct GWindow *gwindow_new(Coord x, Coord y, Coord w, Coord h, DynStr16 *name, struct GObj *child, size_t owner);
int gwindow_regcmd(struct GWindow*, Str16 *cmd, void *ud, GCMDFunc handler);

struct GLabel {
	struct GObj;
	
	DynStr16 *data;
	int owned;
};
struct GLabel *glabel_new(Coord x, Coord y, Coord w, Coord h, DynStr16 *data, int owned);

struct GListChild {
	DynStr16 *text;
	void *ud;
	uint8_t selected;
};
struct GList {
	struct GObj;
	
	size_t childCount;
	struct GListChild *children;
	
	size_t cursor;
};
struct GList *glist_new(Coord x, Coord y, Coord w, Coord h);
int glist_add(struct GList*, DynStr16*, void *ud);
void glist_clear(struct GList*);

DynStr16 *g_dynstr16_new(uint16_t cap, uint16_t len, const void *data);
DynStr16 *g_dynstr16_append(DynStr16*, uint16_t len, const void *data);
void g_dynstr16_free(DynStr16*);

Str16 *g_str16_new(uint16_t len, const void *data);
