#include"canal.h"

#include"vpm.h"
#include"mem.h"
#include"scheduler.h"
#include"ppm.h"

static inline void pr(int c) {
	asm volatile("outb %%al, $0xE9" :: "a"(c) :);
}

static inline void pri(size_t num) {
	char buf[16] = {};
	int i = 16;
	
	do {
		buf[--i] = num % 10;
		num /= 10;
	} while(num);
	
	for(; i < 16; i++) {
		pr(buf[i] + '0');
	}
}

static inline void prcs(const char *str) {
	while(*str) {
		pr(*str);
		str++;
	}
}

static Canal *canals = 0;
static Link *links = 0;

void canal_init() {
	uint32_t cr3;
	asm volatile("movl %%cr3, %0" : "=r"(cr3) :);
	
	canals = vpm_map(cr3, VPM_KERNEL, -1, -1, 4096);
	kmemw1(canals, 0, 4096);
	
	links = vpm_map(cr3, VPM_KERNEL, -1, -1, 4096);
	kmemw1(links, 0xFF, 4096);
}

uint32_t canal_create(uint64_t name, uint16_t hostProcessId) {
	Process *hostProcess = scheduler_find(hostProcessId);
	
	for(int i = 0; i < 4096 / sizeof(Canal); i++) {
		if(canals[i].name == name) return -2;
		
		if(canals[i].name == 0) {
			canals[i].name = name;
			canals[i].hostProcess = hostProcess;
			canals[i].links = NULL;
			
			canals[i].ownedNext = hostProcess->ownedCanal;
			hostProcess->ownedCanal = &canals[i];
			
			return i;
		}
	}
	return -1;
}

uint32_t link_create(uint64_t canalName, uint16_t guestProcessId, uint32_t areaSize) {
	int canalId = -1;
	for(int i = 0; i < 4096 / sizeof(Canal); i++) {
		if(canals[i].name == canalName) {
			canalId = i;
			break;
		}
	}
	
	Process *guest = scheduler_find(guestProcessId);
	
	if(canalId == -1) return -1;
	if(canals[canalId].hostProcess == guest) return -1;
	
	uint32_t hostCR3 = canals[canalId].hostProcess->ctx.cr3;
	uint32_t hostAreaVirtAddr = vpm_find_free(hostCR3, VPM_USER, areaSize);
	
	uint32_t guestCR3 = guest->ctx.cr3;
	uint32_t guestAreaVirtAddr = vpm_find_free(guestCR3, VPM_USER, areaSize);
	
	/* TODO: Fail if areas not found. */
	
	for(int i = 0; i < 4096 / sizeof(Link); i++) {
		if(links[i].canalId == -1) {
			links[i].canalId = canalId;
			links[i].guestProcess = guest;
			links[i].hostVirtAddr = hostAreaVirtAddr;
			links[i].guestVirtAddr = guestAreaVirtAddr;
			links[i].flags = 0;
			
			uint32_t pages = (areaSize + 4095) / 4096;
			
			uint32_t phys = ppm_alloc(pages);
			vpm_map(hostCR3, VPM_USER, hostAreaVirtAddr, phys, areaSize);
			vpm_map(guestCR3, VPM_USER, guestAreaVirtAddr, phys, areaSize);
			
			canals[canalId].hostProcess->eventsRequest++;
			
			links[i].nextInCanal = canals[canalId].links;
			canals[canalId].links = &links[i];
			
			links[i].nextGuestLink = guest->guestLinks;
			guest->guestLinks = &links[i];
			
			return guestAreaVirtAddr;
		}
	}
	
	return -1;
}

uint32_t canal_accept(uint16_t hostProcessId, uint32_t canalId) {
	for(int i = 0; i < 4096 / sizeof(Link); i++) {
		if(links[i].canalId != -1 && canals[links[i].canalId].hostProcess->id == hostProcessId && (canalId == (uint32_t)-1 || links[i].canalId == canalId) && !(links[i].flags & LINK_FLAG_ACCEPTED)) {
			links[i].flags = links[i].flags | LINK_FLAG_ACCEPTED;
			prcs("RETURNING ");
			pri(links[i].hostVirtAddr);
			pr('\n');
			return links[i].hostVirtAddr;
		}
	}
	
	return -1;
}

void link_send_signal_from(uint16_t processId, uint32_t processAreaAddr, uint32_t meta) {
	for(int i = 0; i < 4096 / sizeof(Link); i++) {
		if(links[i].guestProcess->id == processId && links[i].guestVirtAddr == processAreaAddr) {
			if(!(links[i].flags & LINK_FLAG_G2H)) {
				canals[links[i].canalId].hostProcess->eventsSignal++;
				links[i].flags |= LINK_FLAG_G2H;
			}
			
			links[i].metaG2H = meta;
			
			return;
		} else if(canals[links[i].canalId].hostProcess->id == processId && links[i].hostVirtAddr == processAreaAddr) {
			if(!(links[i].flags & LINK_FLAG_H2G)) {
				links[i].guestProcess->eventsSignal++;
				links[i].flags |= LINK_FLAG_H2G;
			}
			
			links[i].metaH2G = meta;
			
			return;
		}
	}
}

uint32_t is_signal_awaiting_to_concerning_link(uint16_t processId, uint32_t area, uint32_t *metaOut, uint32_t *canalId) {
	for(int i = 0; i < 4096 / sizeof(Link); i++) {
		if((links[i].guestProcess->id == processId && (area == -1 || links[i].guestVirtAddr == area) && (links[i].flags & LINK_FLAG_H2G))
			|| (canals[links[i].canalId].hostProcess->id == processId && (area == -1 || area == links[i].hostVirtAddr) && (links[i].flags & LINK_FLAG_G2H))) {
			
			*canalId = links[i].canalId;
			
			return signal_accept(scheduler_find(processId), &links[i], metaOut);
		}
	}
	return -1;
}

uint32_t signal_accept(Process *p, Link *l, uint32_t *meta) {
	p->eventsSignal--;
	
	if(l->guestProcess == p) {
		l->flags &= ~LINK_FLAG_H2G;
		*meta = l->metaH2G;
		return l->guestVirtAddr;
	} else {
		l->flags &= ~LINK_FLAG_G2H;
		*meta = l->metaG2H;
		return l->hostVirtAddr;
	}
}
